/* zipup.c
 Copyright (C) 1990-1996 Mark Adler, Richard B. Wales, Jean-loup Gailly,
 Kai Uwe Rommel, Onno van der Linden and Igor Mandrichenko.
 This version modified By Eric W. Engler for Delphi Zip.
*/

#define NOCPYRT         /* this is not a main module */
#include <ctype.h>
#include "zip.h"

#ifndef UTIL            /* This module contains no code for Zip Utilities */

#include "crypt.h"
#include <windows.h>

/* Use the raw functions for MSDOS and Unix to save on buffer space.
   (This sort of stuff belongs in fileio.c, but oh well.) */

#include "zipup.h"

/* Deflate "internal" global data (currently not in zip.h) */
extern ulg window_size;         /* size of said window */

/* Local data */

  local ulg crc;       /* crc on uncompressed file data */
  local ftype ifile;   /* file to compress */

#ifdef DEBUG
  ulg isize;           /* input file size. global only for debugging */
#else /* !DEBUG */
  local ulg isize;     /* input file size. */
#endif /* ?DEBUG */

/* Local functions */
local int suffixes OF((char *, char *));

/* Return the percentage compression from n to m using only integer
   operations */
int percent(n, m)
ulg n;
ulg m;               /* n is the original size, m is the new size */
{
  if (n > 0xffffffL)            /* If n >= 16M */
  {                             /*  then divide n and m by 256 */
    n += 0x80;  n >>= 8;
    m += 0x80;  m >>= 8;
  }
  return n > m ? (int)(1 + (200 * (n - m)/n)) / 2 : 0;
}

/* Return true if a ends in any of the suffixes in the list s. */
local int suffixes(a, s)
char *a;                /* name to check suffix of */
char *s;                /* list of suffixes separated by : or ; */
{
  int m;                /* true if suffix matches so far */
  char *p;              /* pointer into special */
  char *q;              /* pointer into name a */

  m = 1;
  q = a + strlen(a) - 1;
  for (p = s + strlen(s) - 1; p >= s; p--)
    if (*p == ':' || *p == ';')
      if (m)
        return 1;
      else
      {
        m = 1;
        q = a + strlen(a) - 1;
      }
    else
    {
      m = m && q >= a && case_map(*p) == case_map(*q);
      q--;
    }
  return m;
}

/* Note: a zip "entry" includes a local header (which includes the file
   name), an encryption header if encrypting, the compressed data
   and possibly an extended local header. */

/* Compress the file z->name into the zip entry described by *z and write
   it to the file *y. Encrypt if requested.  Return an error code in the
   ZE_ class.  Also, update tempzn by the number of bytes written. */
int zipup(z, y)
struct zlist far *z;    /* zip entry to compress */
FILE *y;                /* output file */
{
  ztimbuf f_utim;       /* UNIX GMT timestamps, filled by filetime() */
  ulg tim;              /* time returned by filetime() */
  ulg a = 0L;           /* attributes returned by filetime() */
  char *b;              /* malloc'ed file buffer */
  extent k = 0;         /* result of zread */
  int l = 0;            /* true if this file is a symbolic link */
  int m;                /* method for this entry */
  ulg o, p;             /* offsets in zip file */
  long q = -3L;         /* size returned by filetime */
  int r;                /* temporary variable */
  ulg s = 0L;           /* size of compressed data */
  int isdir;            /* set for a directory name */
  int set_type = 0;     /* set if file type (ascii/binary) unknown */
  long fsize = 0L;      /* EWE: filesize for callback */
  HANDLE fin;

  z->nam = strlen(z->zname);
  isdir = z->zname[z->nam-1] == '/';

  if ((tim = filetime(z->name, &a, &q, &f_utim)) == 0 || q < -2L)
    return ZE_OPEN;

  /* q is set to -1 if the input file is a device, -2 for a volume label */
  if (q == -2L) {
     isdir = 1;
     q = 0;
  } else if (isdir != ((a & MSDOS_DIR_ATTR) != 0)) {
     /* don't overwrite a directory with a file and vice-versa */
     return ZE_MISS;
  }
  z->att = (ush)UNKNOWN; /* will be changed later */
  z->atx = 0; /* may be changed by set_extra_field() */

  /* Free the old extra fields which are probably obsolete */
  if (z->ext) {
    free((zvoid *)(z->extra));
  }
  if (z->cext && z->extra != z->cextra) {
    free((zvoid *)(z->cextra));
  }
  z->ext = z->cext = 0;

  window_size = 0L;

  /* Select method based on the suffix and the global method */
  m = special != NULL && suffixes(z->name, special) ? STORE : method;

  if (isdir) {     /* directory */
     ifile = fbad;
     m = STORE;
     q = 0;
  } else {
    /* WARNING: an "fopen" is NOT compatible with GetFileSize in Win32! */
    /* The way we're using CreateFile, it will open existing file for read */
    if ((fin=CreateFile(z->name, GENERIC_READ, FILE_SHARE_READ, NULL,
      OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL))==INVALID_HANDLE_VALUE)
      {
      fsize=-1L;  /* error value (for debugging) */
      return ZE_OPEN;
      }
    else
      {
      /* the file exists - get it's size and close it */
      fsize=GetFileSize(fin, NULL);  // WIN32 API 
      CloseHandle(fin);
      }

    /* Callback:  action, error code, filesize, filename */
    user_callback(1, 0, fsize, z->name);

    /* ============= Here's the open of the input file ================ */
    /* This time it will stay open for a little while :-) */
    if ((ifile = zopen(z->name, fhow)) == fbad)
       return ZE_OPEN;
  }
  z->tim = tim;

  if (l || q == 0)
    m = STORE;
  if (m == BEST)
    m = DEFLATE;

  /* Do not create STORED files with extended local headers if the
   * input size is not known, because such files could not be extracted.
   * So if the zip file is not seekable and the input file is not
   * on disk, obey the -0 option by forcing deflation with stored block.
   * Note however that using "zip -0" as filter is not very useful...
   * ??? to be done.
   */

  /* Fill in header information and write local header to zip file.
   * This header will later be re-written since compressed length and
   * crc are not yet known.
   */

  /* (Assume ext, cext, com, and zname already filled in.) */
  z->vem = z->dosflag ? 20       /* Made under MSDOS by PKZIP 2.0 */
         : OS_CODE + 22;
  /* For a FAT file system, we cheat and pretend that the file
   * was not made on OS2, but under DOS. unzip is confused otherwise.
   */
  z->ver = m == STORE ? 10 : 20;     /* Need PKUNZIP 2.0 except for store */
  z->crc = 0;  /* to be updated later */
  /* Assume first that we will need an extended local header: */
  z->flg = 8;  /* to be updated later */

#ifdef CRYPT
  if (key != NULL) {
    z->flg |= 1;
    /* Since we do not yet know the crc here, we pretend that the crc
     * is the modification time:
     */
    z->crc = z->tim << 16;
  }
#endif /* CRYPT */

  z->lflg = z->flg;
  z->how = m;                             /* may be changed later  */
  z->siz = m == STORE && q >= 0 ? q : 0;  /* will be changed later  */
  z->len = q >= 0 ? q : 0;                /* may be changed later  */
  z->dsk = 0;
  if (z->att == (ush)UNKNOWN) {
      z->att = BINARY;                    /* set sensible value in header */
      set_type = 1;
  }

  /* Attributes from filetime(), flag bits from set_extra_field(): */
  z->atx = z->dosflag ? a & 0xff : a | (z->atx & 0x0000ff00);
  z->off = tempzn;
  if ((r = putlocal(z, y)) != ZE_OK)
    {
    zclose(ifile);
    return r;
    }
  tempzn += 4 + LOCHEAD + z->nam + z->ext;

#ifdef CRYPT
  if (key != NULL) {
    crypthead(key, z->crc, y);
    z->siz += RAND_HEAD_LEN;  /* to be updated later */
    tempzn += RAND_HEAD_LEN;
  }
#endif /* CRYPT */

  if (ferror(y))
    {
    zclose(ifile);
    ziperr(ZE_WRITE);
    }
  o = ftell(y); /* for error checking, ftell can fail on pipes */
  if (ferror(y))
    clearerr(y);

  /* Write stored or deflated file to zip file */
  isize = 0L;
  crc = crc32(0L, (uch *)NULL, 0);

  if (m == DEFLATE) {
     bi_init(y);
     if (set_type)
        z->att = (ush)UNKNOWN; /* will be changed in deflate() */
     ct_init(&z->att, &m);
     lm_init(level, &z->flg);

     /* ========== PERFORM THE DEFLATE ==============*/
     s = deflate();
     if (global_abort_sw)
        {
        zclose(ifile);
        return ZE_ABORT;
        }
  }
  else if (!isdir)
  {
    if ((b = malloc(CBSZ)) == NULL)
       {
       zclose(ifile);
       return ZE_MEM;
       }

    if (l) {
      /* symbolic link processing */
      k = rdsymlnk(z->name, b, CBSZ);

      /* compute crc first because zfwrite 
         will alter the buffer b points to !! */
      crc = crc32(crc, (uch *) b, k);
      if (zfwrite(b, 1, k, y) != k)
      {
        free((zvoid *)b);
        zclose(ifile);
        return ZE_TEMP;
      }
      isize = k;
    }
    else
      while ((k = file_read(b, CBSZ)) > 0 && k != (extent) EOF)
      {
        if (zfwrite(b, 1, k, y) != k)
        {
          free((zvoid *)b);
          zclose(ifile);
          return ZE_TEMP;
        }
      }
    free((zvoid *)b);
    s = isize;
  }
  if (ifile != fbad && zerr(ifile)) {
    zipwarn("could not read input file: ", z->zname);
  }
  if (ifile != fbad)
     zclose(ifile);  /* Close the input file */

  tempzn += s;
  p = tempzn; /* save for future fseek() */


  /* Check input size */
  if (q >= 0 && isize != (ulg)q && !translate_eol)
  {
    Trace((mesg, " i=%ld, q=%ld ", isize, q));
    zipwarn(" file size changed while zipping ", z->name);
  }

  /* Try to rewrite the local header with correct information */
  z->crc = crc;
  z->siz = s;

#ifdef CRYPT
  if (key != NULL)
    z->siz += RAND_HEAD_LEN;
#endif /* CRYPT */

  z->len = isize;
  if (fseek(y, z->off, SEEK_SET)) {
    if (z->how != (ush) m)
       ziperr(ZE_WRITE);
    if (m == STORE && q < 0)
       ziperr(ZE_PARMS);
    if ((r = putextended(z, y)) != ZE_OK)
      return r;
    tempzn += 16L;
    z->flg = z->lflg; /* if flg modified by inflate */
  } else {
     /* seek ok, ftell() should work, check compressed size */
    if (p - o != s) {
      fprintf(mesg, " s=%ld, actual=%ld ", s, p-o);
      ziperr(ZE_LOGIC);
    }
    z->how = m;
    z->ver = m == STORE ? 10 : 20;     /* Need PKUNZIP 2.0 except for store */
    if ((z->flg & 1) == 0)
      z->flg &= ~8; /* clear the extended local header flag */
    z->lflg = z->flg;
    /* rewrite the local header: */
    if ((r = putlocal(z, y)) != ZE_OK)
      return r;
    if (fseek(y, p, SEEK_SET))
      return ZE_READ;
    if ((z->flg & 1) != 0) {
      /* encrypted file, extended header still required */
      if ((r = putextended(z, y)) != ZE_OK)
        return r;
      tempzn += 16L;
    }
  }
  /* Free the local extra field which is no longer needed */
  if (z->ext) {
    if (z->extra != z->cextra)
      free((zvoid *)(z->extra));
    z->ext = 0;
  }

  /* Display statistics */
  if (noisy)
  {
    char tmpbuf[10];

    if (m == DEFLATE)
       strcpy(tmpbuf, "deflated");
    else
       strcpy(tmpbuf, "stored");
    if (verbose)
       {
       fprintf(mesg, "%s  in=%lu,  out=%lu,  %d%%", tmpbuf, isize, s,
                                                       percent(isize, s));
       fflush(mesg);
       }
  }
  return ZE_OK;
}

/* Read a new buffer from the current input file, perform end-of-line
 * translation, and update the crc and input file size.
 * IN assertion: size >= 2 (for end-of-line translation)
 */
int file_read(buf, size)
  char *buf;
  unsigned size;
{
  unsigned len;
  char *b;


  if (translate_eol == 0) {
    len = zread(ifile, buf, size);
    if (len == (unsigned)EOF || len == 0) 
       return (int)len;
  } else if (translate_eol == 1) {
    /* Transform LF to CR LF */
    size >>= 1;
    b = buf+size;
    size = len = zread(ifile, b, size);
    if (len == (unsigned)EOF || len == 0) 
       return (int)len;

    do {
          if ((*buf++ = *b++) == '\n') *(buf-1) = '\r', *buf++ = '\n', len++;
    } while (--size != 0);

    buf -= len;
  } else {
    /* Transform CR LF to LF and suppress final ^Z */
    b = buf;
    size = len = zread(ifile, buf, size-1);
    if (len == (unsigned)EOF || len == 0) 
       return (int)len;
    buf[len] = '\n'; /* I should check if next char is really a \n */

    do {
          if (( *buf++ = *b++) == '\r' && *b == '\n') buf--, len--;
    } while (--size != 0);

    if (len == 0) {
       zread(ifile, buf, 1); len = 1; /* keep single \r if EOF */
    } else {
       buf -= len;
       if (buf[len-1] == CTRLZ) len--; /* suppress final ^Z */
    }
  }
  crc = crc32(crc, (uch *) buf, len);
  isize += (ulg)len;
  return (int)len;
}
#endif /* !UTIL */
